"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});


/**
 * A description of a mutation consumable by mutationWithClientMutationId
 * to create a MutationDefinition 
 *
 * The inputFields and outputFields should not include `clientMutationId: String`,
 * as this will be provided automatically.
 *
 * An input object will be created containing the input fields, and an
 * object will be created containing the output fields.
 *
 * mutateAndGetPayload will receieve an Object with a key for each
 * input field, and it should return an Object with a key for each
 * output field. It may return synchronously, or return a Promise.
 */


/**
 * Copyright (c) 2017, Dirk-Jan Rutten
 * All rights reserved.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 *
 */

var mutationWithClientMutationId = function mutationWithClientMutationId(config) {
  var name = config.name,
      inputFields = config.inputFields,
      outputFields = config.outputFields,
      mutateAndGetPayload = config.mutateAndGetPayload;


  var mutationType = "\n    input " + name + "Input {\n      " + (inputFields ? inputFields : "") + "\n      clientMutationId: String\n    }\n    type " + name + "Payload {\n      " + (outputFields ? outputFields : "") + "\n      clientMutationId: String\n    }\n  ";

  var mutationField = "(input: " + name + "Input!): " + name + "Payload";

  var mutationResolver = function mutationResolver(_, _ref, context, info) {
    var input = _ref.input;

    return Promise.resolve(mutateAndGetPayload(input, context, info)).then(function (payload) {
      payload.clientMutationId = input.clientMutationId;
      return payload;
    });
  };

  return { mutationType: mutationType, mutationField: mutationField, mutationResolver: mutationResolver };
};

/**
 * The definition of the mutation created by mutationWithClientMutationId.
 *
 * The mutationType is the graphQL language type definition 
 * of the input and payload of the mutation.
 *
 * The mutationField contains the arguments and the return type 
 * of the mutation. This should be inserted next to the mutation 
 * name in the root Mutation type.
 *
 * The mutationResolver is the resolver responsible for performing
 * the actual mutation.
 */
exports.mutationWithClientMutationId = mutationWithClientMutationId;